#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import numpy as np
import scipy
import pytest

import resampy


@pytest.mark.parametrize('filt', ['sinc_window', resampy.filters.sinc_window])
@pytest.mark.parametrize('window', [None, scipy.signal.hann])
@pytest.mark.parametrize('num_zeros', [None, 13])
@pytest.mark.parametrize('precision', [None, 9])
@pytest.mark.parametrize('rolloff', [None, 0.925])
def test_filter_sinc(filt, window, num_zeros, precision, rolloff):
    kwargs = dict()
    if window is not None:
        kwargs['window'] = window

    if num_zeros is not None:
        kwargs['num_zeros'] = num_zeros

    if precision is not None:
        kwargs['precision'] = precision

    if rolloff is not None:
        kwargs['rolloff'] = rolloff

    interp, prec, _ = resampy.filters.get_filter(filt, **kwargs)

    if precision in kwargs:
        assert 2**precision == prec


def test_filter_load():
    half_win, precision, _ = resampy.filters.get_filter('kaiser_best')


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_filter_missing():
    resampy.filters.get_filter('bad name')


@pytest.mark.parametrize('sr1, sr2', [(1, 2), (2, 1)])
def test_filter_cache_reset(sr1, sr2):
    x = np.random.randn(100)
    y1 = resampy.resample(x, sr1, sr2, filter='kaiser_fast')

    assert len(resampy.filters.FILTER_CACHE) > 0

    resampy.filters.clear_cache()

    assert len(resampy.filters.FILTER_CACHE) == 0

    y2 = resampy.resample(x, sr1, sr2, filter='kaiser_fast')

    assert np.allclose(y1, y2)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import numpy as np
import pytest

import resampy


def make_tone(freq, sr, duration):
    t = np.arange(int(sr * duration), dtype=np.float64) / sr
    return np.sin(2 * np.pi * freq * t), t


def make_sweep(freq, sr, duration):
    t = np.arange(int(sr * duration), dtype=np.float64) / sr
    x = np.sin(np.cumsum(2 * np.pi * np.logspace(np.log2(2.0 / sr),
                                                 np.log2(float(freq) / sr),
                                                 num=int(duration*sr),
                                                 base=2.0)))
    return x, t


@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
@pytest.mark.parametrize(
    'fil,rms',
    [('sinc_window', 1e-6), ('kaiser_fast', 1e-4), ('kaiser_best', 1e-7)]
)
def test_quality_sine(sr_orig, sr_new, fil, rms):
    FREQ = 512.0
    DURATION = 2.0

    x, _ = make_tone(FREQ, sr_orig, DURATION)
    y, _ = make_tone(FREQ, sr_new, DURATION)
    y_pred = resampy.resample(x, sr_orig, sr_new, filter=fil)

    idx = slice(sr_new // 2, - sr_new//2)

    err = np.mean(np.abs(y[idx] - y_pred[idx]))
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
@pytest.mark.parametrize(
    'fil,rms',
    [('sinc_window', 1e-1), ('kaiser_fast', 1e-1), ('kaiser_best', 1e-1)]
)
def test_quality_sweep(sr_orig, sr_new, fil, rms):
    FREQ = 8192
    DURATION = 5.0
    x, _ = make_sweep(FREQ, sr_orig, DURATION)
    y, _ = make_sweep(FREQ, sr_new, DURATION)

    y_pred = resampy.resample(x, sr_orig, sr_new, filter=fil)

    idx = slice(sr_new // 2, - sr_new//2)

    err = np.mean(np.abs(y[idx] - y_pred[idx]))
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
@pytest.mark.parametrize(
    'fil,rms',
    [('sinc_window', 1e-6), ('kaiser_fast', 1e-4), ('kaiser_best', 1e-7)]
)
def test_resample_nu_quality_sine(sr_orig, sr_new, fil, rms):
    FREQ = 512.0
    DURATION = 2.0

    x, t_in = make_tone(FREQ, sr_orig, DURATION)
    y, t_out = make_tone(FREQ, sr_new, DURATION)

    y_pred = resampy.resample_nu(x, sr_orig, t_out[:-1], filter=fil)

    idx = slice(sr_new // 2, - sr_new//2)

    err = np.mean(np.abs(y[:-1][idx] - y_pred[idx]))
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
@pytest.mark.parametrize(
    'fil,rms',
    [('sinc_window', 1e-1), ('kaiser_fast', 1e-1), ('kaiser_best', 1e-1)]
)
def test_resample_nu_quality_sweep(sr_orig, sr_new, fil, rms):
    FREQ = 8192
    DURATION = 5.0
    x, t_in = make_sweep(FREQ, sr_orig, DURATION)
    y, t_out = make_sweep(FREQ, sr_new, DURATION)
 
    y_pred = resampy.resample_nu(x, sr_orig, t_out[:-1], filter=fil)

    idx = slice(sr_new // 2, - sr_new//2)

    err = np.mean(np.abs(y[:-1][idx] - y_pred[idx]))
    assert err <= rms, '{:g} > {:g}'.format(err, rms)


@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
def test_quality_sine_parallel(sr_orig, sr_new):
    FREQ = 512.0
    DURATION = 2.0
    FILTER = 'sinc_window'
    RTOL = 1e-13

    x, _ = make_tone(FREQ, sr_orig, DURATION)
    y_seq = resampy.resample(x, sr_orig, sr_new, filter=FILTER, parallel=False)
    y_par = resampy.resample(x, sr_orig, sr_new, filter=FILTER, parallel=True)

    np.testing.assert_allclose(y_seq, y_par, rtol=RTOL)
 

@pytest.mark.parametrize('sr_orig,sr_new', [(44100, 22050), (22050, 44100)])
def test_resample_nu_quality_sine_parallel(sr_orig, sr_new):
    FREQ = 512.0
    DURATION = 2.0
    FILTER = 'sinc_window'
    RTOL = 1e-13

    x, _ = make_tone(FREQ, sr_orig, DURATION)
    _, t_out = make_tone(FREQ, sr_new, DURATION)

    y_seq = resampy.resample_nu(x, sr_orig, t_out[:-1], filter=FILTER, parallel=False)
    y_par = resampy.resample_nu(x, sr_orig, t_out[:-1], filter=FILTER, parallel=True)

    np.testing.assert_allclose(y_seq, y_par, rtol=RTOL)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import numpy as np
import scipy.signal
import pytest

import resampy


@pytest.mark.parametrize('axis', [0, 1, 2])
def test_shape(axis):
    sr_orig = 100
    sr_new = sr_orig // 2
    X = np.random.randn(sr_orig, sr_orig, sr_orig)
    Y = resampy.resample(X, sr_orig, sr_new, axis=axis)

    target_shape = list(X.shape)
    target_shape[axis] = target_shape[axis] * sr_new // sr_orig

    assert target_shape == list(Y.shape)


@pytest.mark.parametrize('axis', [0, 1, 2])
def test_resample_nu_shape(axis):
    sr_orig = 100
    sr_new = sr_orig // 2
    X = np.random.randn(sr_orig, sr_orig, sr_orig)
    t = np.arange(X.shape[axis] // 2) / sr_new
    Y = resampy.resample_nu(X, sr_orig, t, axis=axis)

    target_shape = list(X.shape)
    target_shape[axis] = len(t)

    assert target_shape == list(Y.shape)


@pytest.mark.xfail(raises=ValueError, strict=True)
@pytest.mark.parametrize('sr_orig, sr_new', [(100, 0), (100, -1), (0, 100), (-1, 100)])
def test_bad_sr(sr_orig, sr_new):
    x = np.zeros(100)
    resampy.resample(x, sr_orig, sr_new)


@pytest.mark.xfail(raises=ValueError, strict=True)
@pytest.mark.parametrize('sr', [0, -1])
def test_bad_sr_nu(sr):
    x = np.zeros(100)
    t = np.arange(3)
    resampy.resample_nu(x, sr, t)


@pytest.mark.xfail(raises=ValueError, strict=True)
@pytest.mark.parametrize('t', [np.empty(0), np.eye(3)])
def test_bad_time_nu(t):
    x = np.zeros(100)
    resampy.resample_nu(x, 1, t)


@pytest.mark.xfail(raises=ValueError, strict=True)
@pytest.mark.parametrize('rolloff', [-1, 1.5])
def test_bad_rolloff(rolloff):
    x = np.zeros(100)
    resampy.resample(x, 100, 50, filter='sinc_window', rolloff=rolloff)


@pytest.mark.xfail(raises=ValueError, strict=True)
def test_bad_precision():
    x = np.zeros(100)
    resampy.resample(x, 100, 50, filter='sinc_window', precision=-1)


@pytest.mark.xfail(raises=ValueError, strict=True)
def test_bad_num_zeros():
    x = np.zeros(100)
    resampy.resample(x, 100, 50, filter='sinc_window', num_zeros=0)


@pytest.mark.parametrize('dtype', [np.float32, np.float64,
                                   np.complex64, np.complex128])
def test_dtype(dtype):
    x = np.random.randn(100).astype(dtype)

    y = resampy.resample(x, 100, 200)

    assert x.dtype == y.dtype


@pytest.mark.parametrize('dtype', [np.int16, np.int32, np.int64])
def test_dtype_int(dtype):
    x = (32767 * np.random.randn(100)).astype(dtype)

    y = resampy.resample(x, 100, 200)

    assert y.dtype == np.float32


@pytest.mark.parametrize('dtype', [np.int16, np.int32, np.int64])
def test_dtype_int_nu(dtype):
    x = (32767 * np.random.randn(100)).astype(dtype)
    t = np.arange(2 * len(x) - 1) / 2

    y = resampy.resample_nu(x, 1., t)
    assert y.dtype == np.float32


@pytest.mark.parametrize('dtype', [np.float32, np.float64,
                                   np.complex64, np.complex128])
def test_resample_nu_dtype(dtype):
    x = np.random.randn(100).astype(dtype)
    t = np.arange(2 * len(x) - 1) / 2

    y = resampy.resample_nu(x, 1., t)

    assert x.dtype == y.dtype


@pytest.mark.xfail(raises=TypeError)
def test_bad_window():
    x = np.zeros(100)

    resampy.resample(x, 100, 200, filter='sinc_window', window=np.ones(50))


@pytest.mark.xfail(raises=ValueError)
def test_short_signal():

    x = np.zeros(2)
    resampy.resample(x, 4, 1)


@pytest.mark.xfail(raises=ValueError)
def test_resample_nu_short_signal():

    x = np.zeros(2)
    t = np.asarray([])
    resampy.resample_nu(x, 1., t)


def test_good_window():
    sr_orig = 100
    sr_new = 200
    x = np.random.randn(500)
    y = resampy.resample(x, sr_orig, sr_new, filter='sinc_window', window=scipy.signal.blackman)

    assert len(y) == 2 * len(x)


@pytest.mark.parametrize('order', ['C', 'F'])
@pytest.mark.parametrize('shape', [(50,), (10, 50), (10, 25, 50)])
@pytest.mark.parametrize('axis', [0, -1])
def test_contiguity(order, shape, axis):

    x = np.zeros(shape, dtype=np.float64, order=order)
    sr_orig = 1
    sr_new = 2
    y = resampy.resample(x, sr_orig, sr_new, axis=axis)

    assert x.flags['C_CONTIGUOUS'] == y.flags['C_CONTIGUOUS']
    assert x.flags['F_CONTIGUOUS'] == y.flags['F_CONTIGUOUS']


@pytest.mark.parametrize('order', ['C', 'F'])
@pytest.mark.parametrize('shape', [(50,), (10, 50), (10, 25, 50)])
@pytest.mark.parametrize('axis', [0, -1])
def test_resample_nu_contiguity(order, shape, axis):

    x = np.zeros(shape, dtype=np.float64, order=order)
    t = np.arange(x.shape[axis] * 2 - 1) / 2
    y = resampy.resample_nu(x, 1., t, axis=axis)

    assert x.flags['C_CONTIGUOUS'] == y.flags['C_CONTIGUOUS']
    assert x.flags['F_CONTIGUOUS'] == y.flags['F_CONTIGUOUS']


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize('shape', [(50,), (10, 50), (10, 25, 50)])
@pytest.mark.parametrize('axis', [0, -1])
@pytest.mark.parametrize('domain', [(0, 100), (-1, 5)])
def test_resample_nu_domain(shape, axis, domain):

    x = np.zeros(shape, dtype=np.float64)
    t = np.linspace(*domain, num=10, endpoint=True)
    resampy.resample_nu(x, 1., t, axis=axis)


def test_resample_matched():
    x = np.random.randn(100)
    y = resampy.resample(x, 1, 1)

    # All values should match
    assert np.allclose(x, y)
    # y should own the data
    assert y.flags['OWNDATA']
    # x and y are distinct objects
    assert y is not x


def test_resample_axis():
    # derived from https://github.com/bmcfee/resampy/issues/73

    rand_arr = np.abs(np.random.rand(3, 4, 5, 100))

    resampled_arr = resampy.resample(rand_arr, 100, 24, axis=3)

    resampled_t_arr = resampy.resample(np.transpose(rand_arr), 100, 24, axis=0)

    assert np.allclose(resampled_arr, np.transpose(resampled_t_arr))
    assert (resampled_arr**2).sum() > 0


def test_resample_length_rounding():
    # Test for length calculation edge case https://github.com/bmcfee/resampy/issues/111
    x = np.zeros(12499)
    y = resampy.resample(x, 12499, 15001)
    assert len(y) == 15001


def test_resample_rates_fixedprecision():
    # Test for avoiding ill effects of fixed-precision math
    # This should fix a bug in librosa https://github.com/librosa/librosa/issues/1569

    x = np.zeros(720200)
    sr_out = np.int32(4000)
    sr_in = 22050
    resampy.resample(x, sr_in, sr_out)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""Construct resampy filters by Bayesian parameter search"""

import sys
import functools

from argparse import ArgumentParser

import numpy as np
import scipy
import scipy.signal

import optuna

import resampy


def parse_arguments(args):

    parser = ArgumentParser(description=__doc__)

    parser.add_argument(
        "-n",
        "--num-trials",
        dest="n_trials",
        default=1500,
        help="Number of trials",
        type=int,
    )

    parser.add_argument(
        "-z",
        "--num-zeros",
        dest="num_zeros",
        required=True,
        type=int,
        help="Number of zero-crossings",
    )

    parser.add_argument(
        "--min-rolloff",
        dest="min_rolloff",
        default=0.9,
        help="Minimum rolloff frequency (fraction of Nyquist)",
        type=float,
    )

    parser.add_argument(
        "-p",
        "--precision",
        dest="precision",
        default=13,
        help="Precision for filter interpolation",
        type=int,
    )

    parser.add_argument(
        "--min-beta",
        dest="min_beta",
        default=8,
        help="Minimum beta for kaiser filter",
        type=float,
    )

    parser.add_argument(
        "--max-beta",
        dest="max_beta",
        default=20,
        help="Maximum beta for kaiser filter",
        type=float,
    )

    parser.add_argument(
        "--attenuation",
        dest="attenuation",
        default=-120,
        help="Stop-band attenuation",
        type=float,
    )
    parser.add_argument(dest="output_file", type=str, help="Path to store filter")

    return parser.parse_args(args)


def get_window(beta=None, rolloff=None, num_zeros=None):
    """Build the full window from a specification"""
    win = functools.partial(scipy.signal.kaiser, beta=beta)
    fil = resampy.filters.sinc_window(
        num_zeros=num_zeros, precision=1, window=win, rolloff=rolloff
    )[0]

    # The windowed filter
    fil = np.concatenate([fil[-1:0:-1], fil])

    return fil


def _objective(beta, rolloff, attenuation=-120, num_zeros=40):
    """Internal objective function:

    Minimize the mean passband-deviation from unity gain +
    maximum stop-band gain above target attenuation
    """
    fil = get_window(beta, rolloff, num_zeros=num_zeros)

    W, H = scipy.signal.freqz(fil, worN=2048)

    H = np.abs(H)
    H_pass = H[W < 0.5 * np.pi]
    H_stop = H[W >= 0.5 * np.pi]

    err_pass = 20 * np.mean(np.abs(np.log10(H_pass)))
    err_stop = np.max(
        np.abs(np.maximum(20 * np.log10(H_stop), attenuation) - attenuation)
    )
    return err_pass + err_stop


def objective(trial, num_zeros, min_beta, max_beta, min_rolloff, attenuation):
    """Objective wrapper for the optimizer"""

    beta = trial.suggest_float("beta", min_beta, max_beta)
    rolloff = trial.suggest_float("rolloff", min_rolloff, 1.0)

    loss = _objective(beta, rolloff, num_zeros=num_zeros, attenuation=attenuation)

    return loss


if __name__ == "__main__":
    params = parse_arguments(sys.argv[1:])

    optuna.logging.set_verbosity(optuna.logging.WARNING)

    sampler = optuna.samplers.TPESampler(seed=20220629)
    study = optuna.create_study(direction="minimize", sampler=sampler)

    func = functools.partial(
        objective,
        num_zeros=params.num_zeros,
        min_beta=params.min_beta,
        max_beta=params.max_beta,
        min_rolloff=params.min_rolloff,
        attenuation=params.attenuation,
    )

    study.optimize(func, n_trials=params.n_trials, show_progress_bar=True)

    print(f"Parameters for {params.output_file}:")
    print("-" * 40)
    print(f"\tbeta        = {study.best_params['beta']:g}")
    print(f"\troll        = {study.best_params['rolloff']:g}")
    print(f"\t# zeros     = {params.num_zeros}")
    print(f"\tprecision   = {params.precision}")
    print(f"\tattenuation = {params.attenuation}")
    print("-" * 40)
    print(f"Objective value: {study.best_value:g}")

    window = functools.partial(scipy.signal.kaiser, beta=study.best_params["beta"])
    half_win, precision, roll = resampy.filters.sinc_window(
        num_zeros=params.num_zeros,
        precision=params.precision,
        window=window,
        rolloff=study.best_params["rolloff"],
    )

    np.savez(
        params.output_file, half_window=half_win, precision=precision, rolloff=roll
    )
